import os
import yaml


def load_config(filename=None):
    """Load sorunlib config file, using SORUNLIB_CONFIG by default.

    Args:
        filename (str): Path to sorunlib config file

    Returns:
        dict

    """
    if filename is None:
        assert (os.getenv('SORUNLIB_CONFIG') is not None)
        filename = os.getenv('SORUNLIB_CONFIG')

    with open(filename, 'rb') as f:
        return yaml.safe_load(f)



import time

import sorunlib as run
from sorunlib._internal import check_response, check_running

EL_DIFF_THRESHOLD = 0.5  # deg diff from target that its ok to run calibration
BORESIGHT_DIFF_THRESHOLD = 0.5  # deg
AGENT_TIMEDIFF_THRESHOLD = 5  # sec
OP_TIMEOUT = 60


# Internal Helper Functions
def _check_process_data(process, last_timestamp):
    """Check the latest timestamp from a process' session.data is recent enough.

    Args:
        process (str): Descriptive name, used in the error message.
        last_timestamp (float): Latest timestamp from the session.data.

    Raises:
        RuntimeError: When the last timestamp was more than
            AGENT_TIMEDIFF_THRESHOLD old.

    """
    try:
        assert ((time.time() - last_timestamp) < AGENT_TIMEDIFF_THRESHOLD)
    except AssertionError:
        error = f"{process} has no updated data. Cannot proceed with " + \
            "wiregrid calibration. Aborting."
        raise RuntimeError(error)


def _verify_temp_response(response, sensor, min_temp):
    """Check temperature above minimum temperation for safe operation of wiregrid.

    Args:
        response (ocs.ocs_client.OCSReply): Client response from the labjack
            acq process.
        sensor (str): Sensor name in the labjack session data, i.e. 'AIN0'.
        min_temp (int, float): Minimum temperature for safe operation in C.

    Raises:
        RuntimeError: When the temperature of sensor below ``min_temp``.

    """
    temp = response.session['data']['data'][sensor]
    try:
        assert (temp > min_temp)
    except AssertionError:
        error = f"Sensor {sensor} reads {temp} C, which is below minimum " + \
            f"temperature of {min_temp} C. Cannot proceed with wiregrid " + \
            "calibration. Aborting"
        raise RuntimeError(error)


def _check_zenith():
    """Checks to see if the telescope is currently pointing at zenith.

    Returns:
        bool: True if the telescope is at zenith, otherwise returns False.

    """
    acu = run.CLIENTS['acu']
    resp = acu.monitor.status()
    el = resp.session['data']['StatusDetailed']['Elevation current position']

    zenith = False
    if (abs(el - 90) < EL_DIFF_THRESHOLD):
        zenith = True

    return zenith


# Calibration Functions
def _check_telescope_position(elevation_check=True, boresight_check=True):
    # Get current telescope position
    acu = run.CLIENTS['acu']
    resp = acu.monitor.status()
    az = resp.session['data']['StatusDetailed']['Azimuth current position']
    el = resp.session['data']['StatusDetailed']['Elevation current position']
    boresight = resp.session['data']['StatusDetailed']['Boresight current position']

    # Check appropriate elevation
    if elevation_check:
        try:
            assert (el > 48 - EL_DIFF_THRESHOLD)
        except AssertionError:
            error = "Telescope not at > 48 deg elevation. Cannot proceed with " + \
                    f"wiregrid calibration in current position ({az}, {el}). " + \
                    "Aborting."
            raise RuntimeError(error)

    # Check boresight angle
    if boresight_check:
        try:
            assert (abs(boresight - 0) < BORESIGHT_DIFF_THRESHOLD)
        except AssertionError:
            error = "Telescope not at 0 deg boresight. Cannot proceed with " + \
                    f"wiregrid calibration in current position ({boresight}). " + \
                    "Aborting."
            raise RuntimeError(error)


def _configure_power(continuous):
    """Configure the KIKUSUI power supply for rotation based on rotation type.

    Args:
        continuous (bool): Configure for continuous rotation or not.

    """
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    resp = kikusui.set_v(volt=12)
    check_response(kikusui, resp)

    cfg = run.config.load_config()
    current = cfg.get('wiregrid_motor_current', 3.0)

    resp = kikusui.set_c(current=current)
    check_response(kikusui, resp)


def _rotate_continuously(duration):
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    # Start rotation
    resp = kikusui.set_on()
    check_response(kikusui, resp)

    # Wait
    time.sleep(duration)

    # Stop rotation
    resp = kikusui.set_off()
    check_response(kikusui, resp)


def _check_motor_on():
    actuator = run.CLIENTS['wiregrid']['actuator']

    # Check motor is on
    resp = actuator.acq.status()
    if resp.session['data']['fields']['motor'] == 1:
        print("Wiregrid motor already on.")
    # Turn on motor if needed
    elif resp.session['data']['fields']['motor'] == 0:
        resp = actuator.motor_on()
        check_response(actuator, resp)
    else:
        raise RuntimeError("Motor state unknown. Aborting...")


def _check_agents_online():
    """Check OCS agents related to the wiregrid are running."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    kikusui = run.CLIENTS['wiregrid']['kikusui']
    encoder = run.CLIENTS['wiregrid']['encoder']
    labjack = run.CLIENTS['wiregrid']['labjack']

    # wiregrid_actuator
    resp = actuator.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(actuator, resp)
    _check_process_data("Actuator agent", last_timestamp)

    # wiregrid_kikusui
    resp = kikusui.IV_acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(kikusui, resp)
    _check_process_data("Kikusui agent", last_timestamp)

    # wiregrid_encoder
    resp = encoder.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(encoder, resp)
    _check_process_data("Encoder agent", last_timestamp)

    # labjack
    resp = labjack.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(labjack, resp)
    _check_process_data("Labjack agent", last_timestamp)

    # encoder data stream
    resp = encoder.acq.status()
    last_timestamp = resp.session['data']['fields']['encoder_data']['last_updated']
    _check_process_data("Encoder BBB", last_timestamp)


def _check_temperature_sensors():
    """Check temperatures within safe range."""
    labjack = run.CLIENTS['wiregrid']['labjack']
    resp = labjack.acq.status()
    # AIN0 > -10 C && AIN1 > -10 C (On the wiregrid)
    _verify_temp_response(resp, 'AIN0C', -10)
    _verify_temp_response(resp, 'AIN1C', -10)
    # AIN2 > 0 C (Inside the electronics enclosure for the gridloader)
    _verify_temp_response(resp, 'AIN2C', 0)


def _check_wiregrid_position():
    """Check the wiregrid position.

    Returns:
        str: The wiregrid position, either 'inside' or 'outside'.

    Raises:
        RuntimeError: When the wiregrid position is unknown.

    """
    actuator = run.CLIENTS['wiregrid']['actuator']
    resp = actuator.acq.status()
    position = resp.session['data']['fields']['position']
    if position not in ['inside', 'outside']:
        raise RuntimeError("The wiregrid position is unknown. Aborting...")
    return position


# Public API
def insert():
    """Insert the wiregrid."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    resp = actuator.insert()
    try:
        check_response(actuator, resp)
    except RuntimeError as e:
        error = "Wiregrid insertion failed. Please inspect wiregrid before " + \
                "continuing observations.\n" + str(e)
        raise RuntimeError(error)


def eject():
    """Eject the wiregrid."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    resp = actuator.eject()
    try:
        check_response(actuator, resp)
    except RuntimeError as e:
        error = "Wiregrid eject failed. Please inspect wiregrid before " + \
                "continuing observations.\n" + str(e)
        raise RuntimeError(error)


def rotate(continuous, duration=30, num_laps=1, stopped_time=10.):
    """Rotate the wiregrid.

    Rotation is either continuously for ``duration`` seconds or stepwise for
    ``num_laps``, stopping every 22.5 degrees for ``stopped_time`` seconds.

    This function sets the Kikusui power settings appropriately based on the
    rotation type.

    Args:
        continuous (bool): Rotate the grid continuously if True or not if
            False.
        duration (int, float): Amount of time in seconds to rotate if rotating
            continuously. Defaults to 30 seconds.
        num_laps (int): Number of revolutions if rotating stepwise (i.e.
            ``continuous = False``). Defaults to 1.
        stopped_time (float): Duration of each 22.5 deg step in seconds if
            rotating stepwise. Defaults to 10 seconds.

    """
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    # Configure power settings based on rotation type
    _configure_power(continuous)

    if continuous:
        _rotate_continuously(duration)
    else:
        resp = kikusui.stepwise_rotation(num_laps=num_laps,
                                         stopped_time=stopped_time)
        check_response(kikusui, resp)


def calibrate(continuous=False, elevation_check=True, boresight_check=True,
              temperature_check=True):
    """Run a wiregrid calibration.

    Args:
        continuous (bool): Calibration by continuous rotation or not.
            Default is False, in which the wiregrid rotates step-wisely.
        elevation_check (bool): Check the elevation angle is in an appropriate
            range before the calibration or not. Default is True.
        boresight_check (bool): Check the boresight angle is in an appropriate
            range before the calibration or not. Default is True.
        temperature_check (bool): Check the temperature of various components
            are within operational limits before the calibration or not.
            Default is True.

    """
    try:
        _check_telescope_position(elevation_check=elevation_check,
                                  boresight_check=boresight_check)
        _check_agents_online()
        if temperature_check:
            _check_temperature_sensors()
        _check_motor_on()

        # Rotate for reference before insertion
        rotate(continuous=True, duration=10)

        # Enable SMuRF streams
        if continuous:
            rotation = 'wg_continuous'
        else:
            rotation = 'wg_stepwise'
        if _check_zenith():
            el_tag = ', wg_el90'
        else:
            el_tag = ''
        run.smurf.stream('on', tag=f'wiregrid, {rotation}{el_tag}', subtype='cal')

        # Insert the wiregrid
        insert()

        # Rotate the wiregrid
        rotate(continuous)

        # Eject the wiregrid
        eject()
    finally:
        # Stop SMuRF streams
        run.smurf.stream('off')


def time_constant(num_repeats=1):
    """
    Run a wiregrid time constant measurement.

    Args:
        num_repeats (int): Number of repeats. Default is 1.
            If this is odd, the HWP direction will be changed to the opposite
            of the initial direction. If this is even, the HWP direction will be
            the same as the initial direction.

    """
    # Check the number of repeats
    if num_repeats < 1 or not isinstance(num_repeats, int):
        error = "The ``num_repeats`` should be int and larger than 0."
        raise RuntimeError(error)

    _check_agents_online()
    _check_motor_on()
    _check_telescope_position(elevation_check=True, boresight_check=False)
    _check_wiregrid_position()
    if _check_wiregrid_position() == 'inside':
        error = "The wiregrid is already inserted before the wiregrid time " + \
                "constant measurement. Please inspect wiregrid and HWP " + \
                "before continuing observations."
        raise RuntimeError(error)

    if _check_zenith():
        el_tag = ', wg_el90'
    else:
        el_tag = ''

    # Check the current HWP direction
    try:
        current_hwp_direction = run.hwp._get_direction()  # 'cw' or 'ccw'
    except RuntimeError as e:
        error = "Wiregrid time constant measurment has failed " + \
                "due to a failure in getting the HWP direction.\n" + str(e)
        raise RuntimeError(error)

    # Rotate to get encoder reference before insertion
    rotate(continuous=True, duration=10)

    # Bias step (the wire grid is off the window)
    bs_tag = 'wiregrid, wg_time_constant, wg_ejected, ' + \
             f'hwp_{current_hwp_direction}' + el_tag
    run.smurf.bias_step(tag=bs_tag, concurrent=True)
    time.sleep(5)

    # Insert the wiregrid while streaming
    try:
        stream_tag = 'wiregrid, wg_time_constant, wg_inserting, ' + \
                     f'hwp_{current_hwp_direction}' + el_tag
        run.smurf.stream('on', tag=stream_tag, subtype='cal')
        insert()
        time.sleep(5)
    finally:
        run.smurf.stream('off')

    for i in range(num_repeats):
        if current_hwp_direction == 'ccw':
            target_hwp_direction = 'cw'
        elif current_hwp_direction == 'cw':
            target_hwp_direction = 'ccw'

        # Bias step (the wire grid is on the window)
        # Before stopping the HWP
        bs_tag = 'wiregrid, wg_time_constant, wg_inserted, ' + \
                 f'hwp_{current_hwp_direction}' + el_tag
        run.smurf.bias_step(tag=bs_tag, concurrent=True)

        # Run stepwise rotation before stopping the HWP
        try:
            stream_tag = 'wiregrid, wg_time_constant, ' + \
                         f'wg_stepwise, hwp_{current_hwp_direction}' + \
                         el_tag
            run.smurf.stream('on', tag=stream_tag, subtype='cal')
            # Run stepwise rotation
            rotate(continuous=False)
        finally:
            run.smurf.stream('off')

        # Stop the HWP while streaming
        try:
            stream_tag = 'wiregrid, wg_time_constant, ' + \
                         f'hwp_change_{current_hwp_direction}_to_stop' + el_tag
            run.smurf.stream('on', tag=stream_tag, subtype='cal')
            run.hwp.stop(active=True)
        finally:
            run.smurf.stream('off')

        # Reverse the HWP while streaming
        try:
            stream_tag = 'wiregrid, wg_time_constant, ' + \
                         f'hwp_change_stop_to_{target_hwp_direction}' + el_tag
            run.smurf.stream('on', tag=stream_tag, subtype='cal')
            # Note: This is hardcoding the correspondance between direction and
            # the sign of the frequency, which is subject to change depending
            # on the hardware/agent configuration. This should be removed in
            # the future, if possible.
            if target_hwp_direction == 'ccw':
                run.hwp.set_freq(freq=2.0)
            elif target_hwp_direction == 'cw':
                run.hwp.set_freq(freq=-2.0)
            current_hwp_direction = target_hwp_direction
        finally:
            run.smurf.stream('off')

    # Run stepwise rotation after changing the HWP rotation
    try:
        stream_tag = 'wiregrid, wg_time_constant, ' + \
                     f'wg_stepwise, hwp_{current_hwp_direction}' + \
                     el_tag
        run.smurf.stream('on', tag=stream_tag, subtype='cal')
        # Run stepwise rotation
        rotate(continuous=False)
    finally:
        run.smurf.stream('off')

    # Bias step (the wire grid is on the window)
    # After changing the HWP rotation
    bs_tag = 'wiregrid, wg_time_constant, wg_inserted, ' + \
             f'hwp_{current_hwp_direction}' + el_tag
    run.smurf.bias_step(tag=bs_tag, concurrent=True)
    time.sleep(5)

    # Eject the wiregrid while streaming
    try:
        stream_tag = 'wiregrid, wg_time_constant, wg_ejecting, ' + \
                     f'hwp_{current_hwp_direction}' + el_tag
        run.smurf.stream('on', tag=stream_tag, subtype='cal')
        eject()
        time.sleep(5)
    finally:
        run.smurf.stream('off')

    # Bias step (the wire grid is off the window)
    bs_tag = 'wiregrid, wg_time_constant, wg_ejected, ' + \
             f'hwp_{current_hwp_direction}' + el_tag
    run.smurf.bias_step(tag=bs_tag, concurrent=True)



import time
import sorunlib as run
from sorunlib._internal import check_response

ID_SHUTTER = 1


def _open_shutter():
    """Open the shutter of the stimulator"""
    ds = run.CLIENTS['stimulator']['ds378']
    resp = ds.set_relay(relay_number=ID_SHUTTER, on_off=1)
    check_response(ds, resp)

    time.sleep(1)


def _close_shutter():
    """Close the shutter of the stimulator"""
    ds = run.CLIENTS['stimulator']['ds378']
    resp = ds.set_relay(relay_number=ID_SHUTTER, on_off=0)
    check_response(ds, resp)

    time.sleep(1)


def _setup():
    # Open shutter
    _open_shutter()

    # Acceleration / Decceleration configuration
    blh = run.CLIENTS['stimulator']['blh']
    resp = blh.set_value(accl_time=10, decl_time=10)
    check_response(blh, resp)


def _stop():
    blh = run.CLIENTS['stimulator']['blh']

    # Stop rotation
    resp = blh.stop_rotation()
    check_response(blh, resp)
    time.sleep(10)

    # Close shutter
    _close_shutter()


def calibrate_tau(duration_step=10,
                  speeds_rpm=[225, 495, 945, 1395, 1845, 2205],
                  forward=True, do_setup=True, stop=True,
                  downsample_factor=8, filter_order=4, filter_cutoff=300):
    """Time constant calibration using the stimulator.

    Parameters
    ----------
    duration_step : float, optional
        Duration of each step of time constant measurement in sec, default to 10 sec.
    speeds_rpm : list of float, optional
        List of chopper rotation speed in RPM for each step. Defaults to [225, 495, 945, 1395, 1845, 2205].
    forward : bool, optional
        Chopper rotation direction. If True, the chopper rotates clockwise
        when viewed from the receiver toward the stimulator. Defaults to True.
    do_setup : bool, optional
        Do initial setup (i.e. open shutter, set acceleration, start rotation) if True. Defaults to True.
    stop : bool, optional
        Stop the rotation and close the shutter if True. Defaults to True.
    downsample_factor : int, optional
        Downsample factor for SMuRF. Defaults to 8.
    filter_order : int, optional
        Order of the downsample filter for SMuRF. Defaults to 4.
    filter_cutoff : float, optional
        The cutoff frequency in Hz for the downsample filter for SMuRF. Defaults to 300.
    """

    blh = run.CLIENTS['stimulator']['blh']

    try:
        run.smurf.stream('on', tag='stimulator, time_constant', subtype='cal',
                         filter_order=filter_order, filter_cutoff=filter_cutoff,
                         downsample_factor=downsample_factor)

        if do_setup:
            _setup()
            # Rotation setting
            resp = blh.set_value(speed=speeds_rpm[0])
            check_response(blh, resp)

            resp = blh.start_rotation(forward=forward)
            check_response(blh, resp)

            speeds_rpm = speeds_rpm[1:]

            # First data point
            time.sleep(duration_step)

        for speed_rpm in speeds_rpm:
            resp = blh.set_value(speed=speed_rpm)
            check_response(blh, resp)

            time.sleep(duration_step)
    finally:
        try:
            run.smurf.stream('off')
        except RuntimeError as e:
            print(f"Caught error while shutting down SMuRF streams: {e}")

        if stop:
            _stop()


def calibrate_gain(duration=60, speed_rpm=90,
                   forward=True, do_setup=True, stop=True):
    """Gain calibration with the stimulator

    Parameters
    ----------
    duration : float, optional
        Duration of the gain calibration in sec, default to 60 sec.
    speed_rpm : float, optional
        Rotation speed of the chopper wheel in RPM, default to 90 RPM.
    forward : bool, optional
        Chopper rotation direction. If True, the chopper rotates clockwise
        when viewed from the receiver toward the stimulator. Defaults to True.
    do_setup : bool, optional
        Do initial setup (i.e. open shutter, set acceleration, start rotation) if True.
        Defaults to True.
    stop : bool, optional
        Stop the rotation and close the shutter if True. Defaults to True.
    """
    blh = run.CLIENTS['stimulator']['blh']

    try:
        resp = blh.set_value(speed=speed_rpm)
        check_response(blh, resp)

        if do_setup:
            _setup()
            # Rotation setting
            resp = blh.start_rotation(forward=forward)
            check_response(blh, resp)

        # Sleep for rotation stabilization
        time.sleep(10)

        run.smurf.stream('on', tag='stimulator, gain', subtype='cal')

        # Data taking
        time.sleep(duration)
    finally:
        try:
            run.smurf.stream('off')
        except RuntimeError as e:
            print(f"Caught error while shutting down SMuRF streams: {e}")

        if stop:
            _stop()



import datetime as dt
import time

import sorunlib as run

from sorunlib.commands import _timestamp_to_utc_datetime
from sorunlib._internal import check_response, check_started, monitor_process


OP_TIMEOUT = 60


def _stop_smurfs():
    # Stop SMuRF streams
    try:
        run.smurf.stream('off')
    except RuntimeError as e:
        print(f"Caught error while shutting down SMuRF streams: {e}")


def _stop_scan():
    acu = run.CLIENTS['acu']

    print("Stopping scan.")
    _stop_smurfs()

    # Stop motion
    acu.generate_scan.stop()
    resp = acu.generate_scan.wait(timeout=OP_TIMEOUT)
    check_response(acu, resp)
    print("Scan finished.")


def scan(description, stop_time, width, az_drift=0, tag=None, subtype=None,
         min_duration=None):
    """Run a constant elevation scan, collecting detector data.

    Args:
        description (str): Description of the field/object being scanned.
        stop_time (str): Time in ISO format to scan until, i.e.
            "2022-06-21T15:58:00"
        width (float): Scan width in azimuth. The scan will start at the
            current position and move in the positive azimuth direction.
        az_drift (float): Drift velocity in deg/s, causing scan extrema to move
            accordingly.
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation. Passed through to the smurf stream command.
        subtype (str, optional): Operation subtype used to tag the stream.
        min_duration (float, optional): Minimum duration required to scan,
            specified in seconds. If not enough time exists between now and the
            ``stop_time`` the scan is not executed. Defaults to None.

    """
    now = dt.datetime.now(dt.timezone.utc)
    scan_stop = _timestamp_to_utc_datetime(stop_time)

    # Check stop time has not already passed
    if now > scan_stop:
        return

    # Check there is enough time to perform scan
    if min_duration is not None:
        start_by_time = scan_stop - dt.timedelta(seconds=min_duration)
        if now > start_by_time:
            return

    acu = run.CLIENTS['acu']

    # Enable SMuRF streams
    run.smurf.stream('on', subtype=subtype, tag=tag)

    try:
        # Grab current telescope position
        resp = acu.monitor.status()
        az = resp.session['data']['StatusDetailed']['Azimuth current position']
        el = resp.session['data']['StatusDetailed']['Elevation current position']

        # Start telescope motion
        # az_speed and az_accel assumed from ACU defaults
        # Can be modified by acu.set_scan_params()
        resp = acu.generate_scan.start(az_endpoint1=az,
                                       az_endpoint2=az + width,
                                       el_endpoint1=el,
                                       el_endpoint2=el,
                                       el_speed=0,
                                       az_drift=az_drift)
        check_started(acu, resp)

        # Wait until stop time
        monitor_process(acu, 'generate_scan', stop_time)
    finally:
        _stop_scan()


def el_nod(el1, el2, num=5, pause=5):
    """Perform a set of elevation nods.

    Elevation nods will be peformed at the current azimuth, and will start from
    and return to the current elevation. The nod first moves to ``el1``,
    pauses, then moves to ``el2``, pauses, and then repeats for the
    specified number of iterations.

    Args:
        el1 (float): First elevation to move to during the nod.
        el2 (float): Second elevation to move to during the nod.
        num (int): Number of nods to peform. Defaults to 5.
        pause (float): Length of pause, in seconds, at each elevation. Defaults
            to 5 seconds.

    """
    acu = run.CLIENTS['acu']

    # Enable SMuRF streams
    run.smurf.stream('on', subtype='cal', tag='el_nods')

    try:
        # Grab current telescope position
        resp = acu.monitor.status()
        init_az = resp.session['data']['StatusDetailed']['Azimuth current position']
        init_el = resp.session['data']['StatusDetailed']['Elevation current position']

        # Perform nods
        for x in range(num):
            run.acu.move_to(az=init_az, el=el1)
            time.sleep(pause)
            run.acu.move_to(az=init_az, el=el2)
            time.sleep(pause)
        else:
            # Return to initial position
            run.acu.move_to(az=init_az, el=init_el)
    finally:
        _stop_smurfs()



from . import acu, hwp, seq, smurf, stimulator, wiregrid

from .commands import wait_until
from .util import create_clients

CLIENTS = None


def initialize(test_mode=False):
    """Initialize global clients list.

    Args:
        test_mode (bool): Operate in 'test mode'. Use this to find Agents that
            are meant to stand in for real agents while testing, i.e.
            SmurfFileEmulators instead of PysmurfControllers.
    """
    global CLIENTS
    CLIENTS = create_clients(test_mode=test_mode)


__all__ = ["acu",
           "hwp",
           "seq",
           "smurf",
           "stimulator",
           "wiregrid",
           "wait_until",
           "initialize"]

# Define the variable '__version__':
# This has the closest behavior to versioneer that I could find
# https://github.com/maresb/hatch-vcs-footgun-example
try:
    # If setuptools_scm is installed (e.g. in a development environment with
    # an editable install), then use it to determine the version dynamically.
    from setuptools_scm import get_version

    # This will fail with LookupError if the package is not installed in
    # editable mode or if Git is not installed.
    __version__ = get_version(root="..", relative_to=__file__, version_scheme="no-guess-dev")
except (ImportError, LookupError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        from sorunlib._version import __version__  # noqa: F401
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package, so complain.
        raise RuntimeError(
            "sorunlib is not correctly installed. "
            "Please install it with pip."
        )



import datetime as dt

import sorunlib as run

from sorunlib.commands import _timestamp_to_utc_datetime
from sorunlib._internal import check_response


def move_to(az, el):
    """Move telescope to specified coordinates.

    Args:
        az (float): destination angle for the azimuthal axis
        el (float): destination angle for the elevation axis

    """
    acu = run.CLIENTS['acu']
    resp = acu.go_to(az=az, el=el)
    check_response(acu, resp)


def move_to_target(az, el, start_time, stop_time, drift):
    """Move telescope to a given drifting target.

    Args:
        az (float): Destination angle for the azimuthal axis.
        el (float): Destination angle for the elevation axis.
        start_time (str): Time in ISO format that target is at the given
            coordinates, i.e. "2024-09-22T07:13:34.416664+00:00".
        stop_time (str): Time in ISO format that the target is no longer
            available to scan, i.e. "2024-09-22T08:42:16.343049+00:00".
        drift (float): Azimuthal drift rate of the target in degrees per
            second. Used to adjust ``az`` if the move occurs after
            ``start_time`` but before ``stop_time``.

    """
    start = _timestamp_to_utc_datetime(start_time)
    stop = _timestamp_to_utc_datetime(stop_time)
    now = dt.datetime.now(dt.timezone.utc)

    if now > start and now < stop:
        az = az + drift * (now - start).total_seconds()
        print(f"Target has drifted since {start_time}. Moving to ({az}, {el}).")

    if now > stop:
        return

    move_to(az, el)


def set_boresight(target):
    """Move the third axis to a specific target angle.

    Args:
        target (float): destination angle for boresight rotation

    """
    acu = run.CLIENTS['acu']
    resp = acu.set_boresight(target=target)
    check_response(acu, resp)


def set_scan_params(az_speed, az_accel, reset=False):
    """Update the default scan parameters, used during :func:`sorunlib.seq.scan`.

    Args:
        az_speed (float, optional): The azimuth scan speed.
        az_accel (float, optional): The (average) azimuth acceleration at
            turn-around.
        reset (bool, optional): If True, reset all params to default values
            before applying any updates passed explicitly here.

    """
    acu = run.CLIENTS['acu']
    resp = acu.set_scan_params(az_speed=az_speed,
                               az_accel=az_accel,
                               reset=reset)
    check_response(acu, resp)


def set_shutter(action):
    """Control the LAT shutter.

    Args:
        action (str): Shutter action, either 'open' or 'close'.

    """
    acu = run.CLIENTS['acu']
    resp = acu.set_shutter(action=action)
    check_response(acu, resp)



import os
import re

from sorunlib.config import load_config

from ocs import site_config
from ocs.ocs_client import OCSClient
from ocs.client_http import ControlClientError


class CrossbarConnectionError(Exception):
    pass


def _load_site_config(filename=None):
    """Load a site config file, searching for default.yaml in OCS_CONFIG DIR by
    default.

    Args:
        filename (str): Path to OCS site config file

    Returns:
        ocs.site_config.SiteConfig object

    """
    if filename is None:
        assert (os.getenv('OCS_CONFIG_DIR') is not None)
        site = 'default'
        filename = os.path.join(os.getenv('OCS_CONFIG_DIR'),
                                site + '.yaml')
    cfg = site_config.SiteConfig.from_yaml(filename)

    return cfg


def _find_instances(agent_class, host=None, config=None):
    """Find all instances of an Agent Class in a config.

    Args:
        agent_class (str): Agent Class name to search for, must match Agent
            Class defined by an OCS Agent (and thus also defined in the SCF.)
        host (str): Specific host to search on, if None all hosts will be
            searched.
        config (str): Path to the OCS Site Config File. If None the default
            file in OCS_CONFIG_DIR will be used.

    Returns:
        list: List of instance-id's matching the given agent_class.

    """
    cfg = _load_site_config(config)

    if host is not None:
        hosts = {host: cfg.hosts[host]}
    else:
        hosts = cfg.hosts

    instances = []
    for _host, hostcfg in hosts.items():
        for entry in hostcfg.instances:
            if entry['agent-class'] == agent_class:
                instances.append(entry['instance-id'])

    return instances


def _find_active_instances(agent_class):
    """Find all instances of an Agent Class currently online, based on the
    Agents known by the registry.

    Args:
        agent_class (str): Agent Class name to search for, must match Agent
            Class defined by an OCS Agent (and thus also defined in the SCF.)

    Returns:
        str or list: List of instance-id's matching the given agent_class. If
            the list is of length 1, just return the only instance-id.

    """
    cfg = load_config()

    reg_client = _try_client(cfg['registry'])
    _, _, session = reg_client.main.status()

    instances = []
    for entry in session['data'].values():
        if entry['expired']:
            continue

        if entry['agent_class'] == agent_class:
            instance_id = entry['agent_address'].split('.')[-1]
            instances.append(instance_id)

    if len(instances) == 1:
        return instances[0]

    return instances


def _try_client(instanceid):
    """User in place of OCSClient to handle common exceptions."""
    if not instanceid:
        return

    try:
        client = OCSClient(instanceid)
    except ControlClientError as e:
        # crossbar connection error
        if "Failed to connect" in str(e):
            result = re.search(r"(http://[^ ]+)'", str(e))
            crossbar_url = result.group(1)
            error = f"Cannot connect to crossbar server {crossbar_url}. Check your connection."
            raise CrossbarConnectionError(error)
        # likely an agent connection error
        if "no callee registered" in str(e):
            print(f"Could not instantiate OCSClient for '{instanceid}'.")
            return None
        # other errors, i.e. non-200 error codes
        print(f"Unexpected error trying to instantiate OCSClient for '{instanceid}'.")
        raise ControlClientError(e)

    return client


def _create_wiregrid_clients(config=None, sorunlib_config=None):
    """Create all wiregrid related clients for a single platform.

    Args:
        config (str): Path to the OCS Site Config File. If None the default
            file in OCS_CONFIG_DIR will be used.
        sorunlib_config (str): Path to sorunlib config file. If None the path
            from environment variable SORUNLIB_CONFIG is used.

    Returns:
        dict: Dictionary with the keys, 'acutator', 'encoder', 'kikusui', and
            'labjack' with each value being the corresponding OCSClient.

    """
    actuator = _find_active_instances('WiregridActuatorAgent')
    encoder = _find_active_instances('WiregridEncoderAgent')
    kikusui = _find_active_instances('WiregridKikusuiAgent')

    cfg = load_config(filename=sorunlib_config)
    try:
        labjack = cfg['wiregrid']['labjack']
    except KeyError:
        labjack = None

    clients = {'actuator': _try_client(actuator),
               'encoder': _try_client(encoder),
               'kikusui': _try_client(kikusui),
               'labjack': _try_client(labjack)}

    return clients


def _create_stimulator_clients():
    """Create all stimulator related clients.

    Returns:
        dict: Dictionary with the keys, 'blh', 'pcr500ma', and 'ds378'
            with each value being the corresponding OCSClient.

    """
    blh_id = _find_active_instances('BLHAgent')
    pcr500ma_id = _find_active_instances('PCR500MAAgent')
    ds378_id = _find_active_instances('DS378Agent')

    clients = {}

    if blh_id:
        clients['blh'] = _try_client(blh_id)

    if pcr500ma_id:
        clients['pcr500ma'] = _try_client(pcr500ma_id)

    if ds378_id:
        clients['ds378'] = _try_client(ds378_id)

    return clients


def create_clients(config=None, sorunlib_config=None, test_mode=False):
    """Create all clients needed for commanding a single platform.

    Args:
        config (str): Path to the OCS Site Config File. If None the default
            file in OCS_CONFIG_DIR will be used.
        sorunlib_config (str): Path to sorunlib config file. If None the path
            from environment variable SORUNLIB_CONFIG is used.
        test_mode (bool): Operate in 'test mode'. Use this to find Agents that
            are meant to stand in for real agents while testing, i.e.
            SmurfFileEmulators instead of PysmurfControllers.

    Returns:
        dict: Dictionary with the ACU and SMuRF clients needed for commanding
        in the format::

            clients = {'acu': acu_client,
                       'hwp': hwp_supervisor_client,
                       'smurf': [smurf_client1, smurf_client2, smurf_client3],
                       'wiregrid': {'actuator': actuator_client,
                                    'encoder': encoder_client,
                                    'kikusui': kikusui_client,
                                    'labjack': labjack_client}}

    """
    clients = {}

    if test_mode:
        smurf_agent_class = 'SmurfFileEmulator'
    else:
        smurf_agent_class = 'PysmurfController'

    acu_id = _find_active_instances('ACUAgent')
    hwp_id = _find_active_instances('HWPSupervisor')
    smurf_ids = _find_active_instances(smurf_agent_class)

    if acu_id:
        acu_client = _try_client(acu_id)
        clients['acu'] = acu_client
    if hwp_id:
        hwp_client = _try_client(hwp_id)
        clients['hwp'] = hwp_client

    if isinstance(smurf_ids, str):
        # when only a single SMuRF controller online
        smurf_clients = [smurf_ids]
    else:
        # create smurf client list, even if empty
        smurf_clients = [_try_client(x) for x in smurf_ids]
    clients['smurf'] = smurf_clients

    clients['wiregrid'] = _create_wiregrid_clients(
        config=config,
        sorunlib_config=sorunlib_config)

    clients['stimulator'] = _create_stimulator_clients()

    return clients
